import os
import json


def read_example_data(file_name, diff_file=None):
    """
    Reads the data from WILIAM's artifacts to create a valid dictionary
    for loading external objects.

    Parameters
    ----------
    file_name: str
        Name of the json file with the symbols information.
    diff_file: str (Optional)
        Name of the diff file from artifacts, will provide a dictionary
        only with the new ones. By default it is None, all symbols in
        file_name will be checked.

    Returns
    -------
    elements: dict
        Dictionary of elements compatible with the loading function.

    """
    json_dict = json.load(open(file_name,))[0]['symbols']
    root = os.path.split(file_name)[0]

    if diff_file:
        # TODO define which list should be passed
        new = json.load(open(diff_file,))[0]['symbols']['not_found_in_DB']
        elements_to_read = {key: json_dict[key] for key in new}
    else:
        elements_to_read = json_dict

    elements = {}
    for element, data in elements_to_read.items():
        if data['type'] == 'SUBSCRIPT':
            # subscript elements
            elements[element] = {'type': 'SUBSCRIPT',
                                 'values': data['dependencies']}
        elif 'excels' in data:
            # excel elements
            elements[element] = {'type': 'EXTERNAL',
                                 'excel': []}
            for sheet in data['excels']:
                for info in sheet['info']:
                    elements[element]['excel'].append(
                        {'filename': sheet.get('filename'),
                         'sheet': sheet.get('sheet'),
                         'cell': info.get('cellrange'),
                         'x_row_or_cols': info.get('series'),
                         'subs': info.get('indexes'),
                         'root': root}
                    )
    return elements



import json
import pandas as pd
from io import StringIO
from pathlib import Path
import warnings
import pypandoc
from enum import Enum, auto
import itertools
import operator
import logging


class Severity(Enum):
    WARNING = auto()
    ERROR = auto()
    CRITICAL = auto()

    @classmethod
    def list(cls):
        return list(map(lambda c: c.name, cls))


class Check(object):
    def __init__(self, severity, message, check_name):
        self.message = message
        self.check_name = check_name
        self._display_severity = None
        self.display_severity = severity

    def update_value(self, new_severity):
        self.display_severity = new_severity

    @property
    def display_severity(self):
        return self._display_severity

    @display_severity.setter
    def display_severity(self, value):
        if isinstance(value, Severity):
            self._display_severity = value.name
        elif isinstance(value, str) and value in Severity.list():
            self._display_severity = value
        else:
            raise ValueError(f"{self.check_name} only accepts: "
                             f" {', '.join([x.name for x in Severity])}")


class CheckThresholds(Check):
    def __init__(self, thresholds, message, handler, check_name):
        self.handler = handler
        self.check_name = check_name
        self.thresholds = thresholds
        self._default_message = message
        self.message = None

    def update_severity_and_message(self, check_out):
        # using check specific handlers to set the severity and the message
        pct_fails, mssg_elements = self.handler.process_check_returns(
                                                                     check_out)
        self.message = self._default_message.format(*mssg_elements)
        self.display_severity = self.get_severity(pct_fails)

    def update_value(self, new_thresholds):
        self.thresholds = new_thresholds

    @property
    def thresholds(self):
        return self._thresholds

    @thresholds.setter
    def thresholds(self, thresholds):
        if isinstance(thresholds, dict):
            self._thresholds = thresholds
        else:
            raise TypeError(f"{self.check_name} only accepts dictionaries.")

    def get_severity(self, value):
        if value < self.thresholds[Severity.WARNING.name]:
            return Severity.WARNING
        elif value >= self.thresholds[Severity.CRITICAL.name]:
            return Severity.CRITICAL
        else:
            return Severity.ERROR


class MissingValuesHandler(object):
    # this is used when checking missing values in the data
    @staticmethod
    def process_check_returns(check_out):
        pct_fails = 100 * check_out[0] / check_out[1]
        message_elements = [check_out[0], pct_fails]

        return pct_fails, message_elements


class MissingValuesSeriesHandler(object):
    # this one is used when checking the number of missing values in the series
    @staticmethod
    def process_check_returns(check_out):
        pct_fails = 100 * check_out[0] / check_out[1]
        message_elements = [check_out[0], pct_fails]

        return pct_fails, message_elements


class MissingValuesDataHandler(object):
    # this is used when checking missing values in the data, and returns the
    # list of series values for which the missing values where found.
    # this is only used when missing = yes is set for [All] or for [Dataseries]
    # in the config file
    @staticmethod
    def process_check_returns(check_out):
        number_missing = check_out[0][0]  # number fails
        total_values = check_out[0][1]  # total number of data values

        # list of series values where there are missing values for the data
        list_series_values = check_out[1]

        pct_fails = 100 * number_missing / total_values
        message_elements = [number_missing, pct_fails, list_series_values]

        return pct_fails, message_elements


class OutlierValuesHandler(object):
    @staticmethod
    def process_check_returns(check_out):

        if all(map(lambda x: type(x) in [int, float], check_out[0])):
            num_outliers = len(check_out[0])
        else:
            num_outliers = len(list(itertools.chain(*check_out[0])))

        pct_fails = 100 * num_outliers / check_out[1]
        message_elements = [num_outliers,
                            pct_fails, check_out[0]]

        return pct_fails, message_elements


class Report(object):
    # values are pandoc output formats
    valid_report_formats = {
        "html": "html",
        "md": "gfm",
        "pdf": "pdf",
        "docx": "docx"}
    default_report_columns_passing = [
        "original_name",
        "check_name",
        "data_origin",
        "check_pass"  # this is dropped when the report is generated

    ]
    default_report_columns_failing = [
        "original_name",
        "check_name",
        "data_origin",
        "severity",
        "issue_message"
    ]
    default_group_by = ["original_name", "check_name"]
    default_report_file_path = Path.joinpath(Path.cwd(), "report.html")
    default_report_format = "html"
    default_report_config_file = "report-conf.json"
    column_names_map = {
        "original_name": "Parameter name",
        "data_origin": "Data origin",
        "check_name": "Type of check",
        "check_pass": "check_pass",
        "check_out": "Issue details",
        "py_name": "Python-safe parameter name (long)",
        "py_short_name": "Python-safe parameter name (long)",
        "file": "File name",
        "sheet": "Sheet name",
        "transposed": "Transposed",
        "cell": "Cell code",
        "coords": "Coordinates",
        "check_description": "Description of the check",
        "check_target": "Target of the check",
        "check_arg": "Check arguments",
        "x_row_or_col": "Row or column number of the x axis",
        "severity": "Issue severity",
        "issue_message": "Issue description",
    }
    check_names_map = {
        "outlier_values": "Outlier detection",
        "missing_values": "Missing values",
        "missing_values_data": "Missing values in data",
        "missing_values_series": "Missing values in series",
        "series_monotony": "Series monotony",
        "series_range": "Series range",
        "series_increment_type": "Series trend"
    }
    report_messages = {
        "PASSING_HEADER": "<h2>Passing tests</h2>\n<br>\n",
        "FAILING_HEADER": "<h1>Failing tests</h1>\n<br>\n",
        "SUMMARY_HEADER": "<h1>Summary</h1>\n<br>\n",
        "NUM_PARAMS_ANALYZED_BOLD":
        "<b>Number of model parameters analyzed</b>: {}\n",
        "NUM_FAILING_CHECKS_BOLD":
        "<b>Number of failing checks</b>: {} ({:.2f} %)\n",
        "NUM_CHECKS_EVALUATED_BOLD": "<b>Number of checks evaluated</b>: {}\n",
        "NUMBER_ISSUES_PARAMETER": "<h2>Number of issues per parameter</h2>\n",
        "NUMBER_ISSUES_PARAMETER_BY_TYPE":
        "<h2>Number of issues of each severity for each parameter</h2>\n",
        "NUMBER_TYPE_ISSUES": "<h2>Number of issues of each type</h2>\n",
        "NUMBER_SEVERITY_ISSUES": "<h2>Number of issues of each severity</h2>\n",
        "NUMBER_DATA_ORIGIN": "<h2>Number of issues in each data file</h2>\n",
        "WHITESPACE": "<br>\n",
        "NO_FAILING_CHECKS": "All checks passed.\n",
        "NO_PASSING_CHECKS": "All checks failed.\n"
    }

    default_values = {
        "outlier_values": {"WARNING": 20, "ERROR": 30, "CRITICAL": 50},
        "missing_values": {"WARNING": 20, "ERROR": 30, "CRITICAL": 50},
        "missing_values_data": {"WARNING": 20, "ERROR": 30, "CRITICAL": 50},
        "missing_values_series": {"WARNING": 20, "ERROR": 30, "CRITICAL": 50},
        "series_monotony": "CRITICAL",
        "series_range": "ERROR",
        "series_increment_type": "ERROR"}

    # the severity of the issues of these types are accompained with a
    # percentage of fails defines the handler that is needed to process the
    # thresholds
    handler_map = {
        "outlier_values": OutlierValuesHandler,
        "missing_values": MissingValuesHandler,
        "missing_values_data": MissingValuesDataHandler,
        "missing_values_series": MissingValuesSeriesHandler,
        "series_monotony": None,
        "series_range": None,
        "series_increment_type": None}

    default_check_messages = {
     "outlier_values": "The field has {} ({:.1f}%) potential outliers ({})",
     "missing_values": "The field has {} ({:.1f}%) missing values.",
     "missing_values_data": "The field has {} ({:.1f}%) missing values "
                            "corresponding to series values: {}.",
     "missing_values_series": "The field has {} ({:.1f}%) missing series "
                              "values.",
     "series_monotony": "The series is not monotonic.",
     "series_range": "The series values are outside the given range.",
     "series_increment_type":
     "The series trend does not follow the expected trend."}

    def __init__(self, out):

        self.data = out
        self.buffer = StringIO()
        self.verbose = None
        self.report_format = None
        self.report_file_path = None
        self.report_config = None
        self.check_objects = {}

    @property
    def data(self):
        return self._data

    @data.setter
    def data(self, out):
        df = pd.DataFrame(out)

        def generate_data_origin_column(x):
            final_result = []
            for i in range(len(x.file)):
                if x.check_target in ["data", "constant"]:
                    columns = ['file', 'sheet', 'cell']
                    names = ["file", "tab", "cell/range name"]

                elif x.check_target == "series":
                    columns = ['file', 'sheet', 'x_row_or_col']
                    names = ["file", "tab", "series cell/range name"]
                elif x.check_target == "dataseries":
                    columns = ['file', 'sheet', 'x_row_or_col', 'cell']
                    names = ["file", "tab", "series cell/range name",
                             "data cell/range name"]
                else:
                    raise ValueError(
                        f"Invalid check_target '{x.check_target}'")
                string = ", ".join(
                    [
                        f"{name}: {x[col][i]}"
                        for name, col in zip(names, columns)
                    ])
                if string not in final_result:
                    final_result.append(string)

            return " \n".join(final_result)

        if df.empty:
            warnings.warn("Nothing to report. The dataframe is empty.")
        else:
            # adding column with the location of the data in the spreadsheet,
            # and keeping only unique values
            df["data_origin"] = df.apply(
                lambda x: generate_data_origin_column(x),
                axis=1,
                ).astype(str)
        self._data = df

    @property
    def report_config_path(self):
        return self._report_config_path

    @report_config_path.setter
    def report_config_path(self, file_path):
        """
        Checks if the file exists and is a JSON file and sets the
        report_config_file attribute.
        """
        if Path(file_path).is_file():
            if Path(file_path).suffix == ".json":
                self._report_config_path = file_path
            else:
                raise ValueError("Report configuration file must be a JSON "
                                 "file.")
        else:
            raise FileNotFoundError("Report configuration file not found.")

    def __update_report_config(self):
        """
        Updates the check_objects attribute with the values of the
        report_config attribute.
        """

        for check, value in self.report_config.items():
            self.check_objects[check].update_value(value)

    def build_report(self, report_conf=None, verbose=False):

        """
        Generate a StringIO containing the report.
        Every time the method is called, the report configuration is reset to
        defaults, and then updated with the value passed in report_conf.

        Parameters
        -----------
        report_conf: dict or str or Path (optional)
            If a str or Path is passed, it is used as the path to the report
            configuration.
            If a dict is passed, it is used as the report configuration itself.

        verbose: bool (optional)
            If True, the report contains more statistics and passing tests.

        Returns
        --------
        buffer: StringIO
            Containing the report content in html
        """

        # clean the buffer
        self.buffer.truncate(0)
        self.buffer.seek(0)
        self.verbose = verbose

        # neither passing nor failing checks
        if self.data.empty:
            warnings.warn("Nothing to report. The dataframe is empty.")
            return self.buffer

        # generate default check_objects configuration
        self.check_objects = self.__build_default_check_objects()

        if not report_conf:
            warnings.warn("No report configuration provided. Using default.")
            default_config_path = Path(__file__).parent \
                / "default-report-conf.json"
            self.report_config = self.__validate_json(default_config_path)
        else:
            # the user can pass a new configuration file or a dictionary
            if isinstance(report_conf, (str, Path)):
                # check that the report configuration file is valid and load it
                self.report_config = self.__validate_json(report_conf)
            elif isinstance(report_conf, dict):
                self.report_config = report_conf
            else:
                raise ValueError("report_conf must be a str, Path or dict.")

        # check that the configuration defined by the user is valid
        self.__validate_report_configuration()

        # update default self.check_objects with those from the config file
        self.__update_report_config()

        # assign it again to avoid modifying the original dataframe
        # also, the write report may be called multiple times, so we need to
        # keep the original dataframe untouched
        df = self.data.copy(deep=True)
        passing = df.loc[df["check_pass"].isin([True])]
        failing = df.loc[df["check_pass"].isin([False])]

        # renaming columns of the full df and of the group_by columns
        df.rename(columns=self.__class__.column_names_map, inplace=True)
        group_by = [self.__class__.column_names_map[x] for x in
                    self.__class__.default_group_by]

        # generate the report main body (failing checks only)
        if not failing.empty:  # if there are failing checks
            # adds the severity and issue description columns to the failing df
            failing = self.__grade_issue_severity(failing)

        self.__write_report_body(failing, group_by)

        # write the summary of failing tests and display the passing tests
        if self.verbose:
            self.__write_report_summary(df, passing, failing, group_by)

        return self.buffer

    def report_to_file(self, report=None, report_filename=None):
        """
        Writes the report to a file. Uses pandoc to convert the default html
        report to the rest of supported formats.

        Parameters:
        -----------
        report: StringIO (optional)
            StringIO containing the report content in html

        report_filename: str or Path (optional)
            The path of the resulting report file.

        Returns:
        --------
        None

        """
        pandoc_version = pypandoc.get_pandoc_version().split(".")[:2]

        if tuple(int(i) for i in pandoc_version) < (2, 17):
            warnings.warn(f"pandoc version in your system is "
                          f"{'.'.join(pandoc_version)}. "
                          "Please install pandoc version 2.17 or above.")

        # TODO add support for other formats available in pandoc

        # if the report argument is None but the buffer contains the report
        # result, use the buffer, else raise
        if not report:
            if self.buffer.getvalue():  # buffer not empty
                report = self.buffer
                warnings.warn("No report provided. Using that in the buffer.")
            else:
                raise ValueError("No report provided.")

        report = report.getvalue()

        # get report format and path from the argument report_filename
        self.report_format, self.report_file_path = \
            self.__get_report_path_and_suffix(report_filename)

        with open(self.report_file_path, "w") as f:
            if self.report_format == "html":
                f.write(report)
            elif self.report_format == "md":
                pandoc_fmt = self.__class__.valid_report_formats[
                    self.report_format]
                output = pypandoc.convert_text(
                    report.replace('\n', ' ').replace('\r', ''),
                    pandoc_fmt,
                    format="html",
                    extra_args=[]
                )
                f.write(output)
            elif self.report_format == "pdf":
                output = pypandoc.convert_text(
                    report,
                    "pdf",
                    format="html",
                    outputfile=str(self.report_file_path),
                    extra_args=['--pdf-engine', 'xelatex',
                                '-V', 'geometry:landscape',
                                '-V', 'fontsize:6pt']
                )
            elif self.report_format == "docx":
                output = pypandoc.convert_text(
                    report,
                    "docx",
                    format="html",
                    outputfile=str(self.report_file_path),
                    extra_args=[]
                )

    def __build_default_check_objects(self):

        check_objects = {}

        for check_name, defaults in self.__class__.default_values.items():

            if self.__class__.handler_map[check_name] is None:
                # does not need a handler
                check_objects[check_name] = Check(
                    defaults,
                    self.__class__.default_check_messages[check_name],
                    check_name)
            else:
                check_objects[check_name] = CheckThresholds(
                    defaults,
                    self.__class__.default_check_messages[check_name],
                    self.__class__.handler_map[check_name], check_name)

        return check_objects

    def __validate_json(self, report_config_path):
        """
        Validate the report_config_path and load the configuration.

        Parameters
        -----------
        report_config_path: str or Path
            Path to the report configuration file.

        Returns
        --------
        config: dict
            The report configuration.
        """

        if isinstance(report_config_path, str):
            report_config_path = Path(report_config_path)

        self.report_config_path = report_config_path

        try:
            config = json.load(open(self.report_config_path,))
        except ValueError:
            raise ValueError("Report configuration file not correctly"
                             " formatted.")
        return config

    def __validate_report_configuration(self):
        """
        Validate that the report configuration check names are correct.
        Update the default report configuration with the values from the user
        configuration (json).

        Parameters
        ----------
        None

        Returns
        -------
        None

        """

        for check_name, value in self.report_config.items():

            # verify that the check name is correct
            if check_name not in self.handler_map.keys():
                raise ValueError("Invalid key in report configuration"
                                 f" ({check_name}).")
            # verify that the value type is correct
            if not isinstance(value,
                              type(self.__class__.default_values[check_name])):
                raise ValueError(
                    f"{check_name} in the report configuration file should be "
                    f"a {type(self.__class__.default_values[check_name])}.")

            # update the default values with the user configuration
            if isinstance(value, dict):
                # verify that the keys in the Severity thresholds dict are ok
                if not all(map(lambda x: x in Severity.list(), value.keys())):
                    raise ValueError(
                        f"Invalid severity name for {check_name}.")
                # check that all values are integers
                if not all(map(lambda x: type(x) == int, value.values())):
                    raise TypeError(
                        f"{check_name} thresholds must be integers.")

                # check that all values are between 0 and 100
                if not all(map(lambda x: x in range(0, 101, 1),
                               value.values())):
                    raise ValueError(
                        f"{check_name} thresholds must be between 0 and 100.")

                # check that the values are higher for higher severities (e.g.
                # WARNING < CRITICAL < ERROR))
                if list(
                    dict(sorted(value.items(), key=operator.itemgetter(1))
                         ).keys()) != Severity.list():
                    raise ValueError(
                        f"{check_name} thresholds must be defined in "
                        "ascending order, corresponding to the tolerances for:"
                        f" {', '.join(Severity.list())}")

    def __get_report_path_and_suffix(self, report_file_path=None):
        """
        Validates the report path and returns the suffix and the absolute path.

        Parameters
        ----------
        report_filename: str or Path
            The path of the results file passed by the user.

        Returns
        -------
        report_file_path: Path (optional)
            The absolute path of the report file.

        """
        if not report_file_path:
            warnings.warn("No report file path defined. Using default"
                          " (report.html).")
            return "html", Path(__file__).parent.parent / "report.html"

        # check if the report_path passed by the user is a valid path
        if isinstance(report_file_path, (Path, str)):
            if isinstance(report_file_path, str):
                report_file_path = Path(report_file_path)

            if not report_file_path.suffix:
                raise ValueError(
                    "Report path must include the file type suffix "
                    "(e.g. report.html)."
                )
            else:
                suffix = self.__validate_report_file_format(
                    report_file_path.suffix)

                if not report_file_path.is_absolute():
                    logging.info(
                        "Relative path provided for the report file."
                    )
                    return suffix, report_file_path.absolute()
                else:
                    return suffix, report_file_path
        else:
            raise TypeError("Report path must be a string or a Path object.")

    def __validate_report_file_format(self, fmt):
        """
        Validates the format of the json in which the configuration of the
        report is detailed.

        Parameters
        -----------
        fmt: str
            Report format.

        Returns
        --------
        fmt: str
            Validated report format.
        """
        fmt_ = fmt[1:]

        if fmt_ in self.__class__.valid_report_formats:
            return fmt_
        else:
            raise ValueError(
                "The only supported report formats are {}.".format(
                 ", ".join(self.__class__.valid_report_formats.keys()))
                )

    def __grade_issue_severity(self, df_fail):
        """
        Adds severity and issue_message columns to the dataframe of failed
        checks.

        Parameters
        ----------
        df_fail : pandas.DataFrame
            Dataframe with the failed checks

        Returns
        -------
        df_fail : pandas.DataFrame
            Dataframe with the failed checks and the added columns.
        """
        def issue_severity_and_description(check_out, check_object):
            if isinstance(check_object, CheckThresholds):
                check_object.update_severity_and_message(check_out)
            return pd.Series({"severity": check_object.display_severity,
                              "issue_message": check_object.message})

        new_cols = df_fail.apply(
            lambda x: issue_severity_and_description(
                x["check_out"],
                self.check_objects[x["check_name"]]
                ),
            axis=1,
        )
        df_fail = df_fail.merge(new_cols, left_index=True, right_index=True)
        return df_fail

    def __write_report_body(self, failing, group_by):
        """
        Writes the report body.

        Parameters
        ----------
        failing : pandas.DataFrame
            Dataframe with the failed checks.
        group_by : list
            List of columns to group by.

        Returns
        -------
        None

        """
        # Start writing the report contents in the buffer
        self.buffer.write(
            self.__class__.report_messages["FAILING_HEADER"])

        if not failing.empty:
            # Select only the desired output columns, use clean names, and
            # reindex
            failing_simplified = failing[
                self.__class__.default_report_columns_failing
                ].replace(
                    {"check_name": self.__class__.check_names_map}
                    ).rename(
                    columns=self.__class__.column_names_map).set_index(
                        group_by)

            self.__df_to_buffer(failing_simplified)
        else:
            self.buffer.write(
                self.__class__.report_messages["NO_FAILING_CHECKS"])

        self.buffer.write(self.__class__.report_messages["WHITESPACE"])

    def __write_report_summary(self, df, passing, failing, group_by):
        """
        Writes a summary of the results in the buffer.

        Parameters
        ----------
        df : pandas.DataFrame
            Dataframe containing all the results of the checks.
        passing : pandas.DataFrame
            Dataframe containing the passing results of the checks.
        failing : pandas.DataFrame
            Dataframe containing the failing results of the checks.
        group_by : list
            List of columns to group the results by.

        Returns
        -------
        None

        """

        # calculate the statistics of failing tests, if there are any
        if not failing.empty:

            # Select only the desired output columns, use clean names, and
            # reindex

            # [self.__class__.default_report_columns_failing]
            failing = failing.replace(
                    {"check_name": self.__class__.check_names_map}
                    ).rename(
                    columns=self.__class__.column_names_map).set_index(
                        group_by)

            grouped_by_param_name = (
                failing.groupby(
                    by=self.__class__.column_names_map["original_name"]
                    )[self.__class__.column_names_map["severity"]
                      ].count().to_frame(name="Number")
            )

            grouped_by_check_type = (
                failing.groupby(
                    by=self.__class__.column_names_map["check_name"])[
                    self.__class__.column_names_map["severity"]
                ]
                .count()
                .to_frame(name="Number")
            )
            grouped_by_check_type["Percent of all issues"] = \
                100 * grouped_by_check_type["Number"] / \
                grouped_by_check_type["Number"].sum()
            grouped_by_check_type["Percent of all issues"] = \
                grouped_by_check_type["Percent of all issues"].round(
                    decimals=1)

            grouped_by_var_name_and_severity = (
                failing.groupby([
                    self.__class__.column_names_map["original_name"],
                    self.__class__.column_names_map["severity"]])[
                    self.__class__.column_names_map["severity"]].count(
                    ).to_frame(name="Number")).reset_index(
                    ).set_index(
                        self.__class__.column_names_map["original_name"])

            grouped_by_severity = (
                failing.groupby(self.__class__.column_names_map["severity"])[
                    self.__class__.column_names_map["severity"]].count(
                    ).to_frame(name="Number"))
            grouped_by_severity["Percent of all issues"] = \
                100 * grouped_by_severity["Number"]/grouped_by_severity[
                    "Number"].sum()
            grouped_by_severity["Percent of all issues"] = \
                grouped_by_severity["Percent of all issues"].round(decimals=1)

            # all files from which all checks imported data
            # if data for one variable was loaded from two (or more) separate
            # files (very unlikely), but the issue was only in one of them,
            # the other files will also be counted as having an issue.
            all_imported_files = list(itertools.chain.from_iterable(
                [set(row["File name"]) for _, row in failing.iterrows()]))

            # count number of times each file was imported
            a = {x: all_imported_files.count(x) for x in set(
                all_imported_files)}
            # convert to DataFrame
            group_by_data_origin = pd.DataFrame.from_dict(a,
                                                          orient='index',
                                                          columns=["Number"])
            group_by_data_origin.index.name = "File name"

        #######################################################################

        self.buffer.write(self.__class__.report_messages["WHITESPACE"])
        self.buffer.write(self.__class__.report_messages[
            "SUMMARY_HEADER"])

        # Lines with totals
        self.buffer.write(
            self.__class__.report_messages[
                "NUM_PARAMS_ANALYZED_BOLD"].format(
                    len(df[self.__class__.column_names_map[
                        "original_name"]].unique())
            )
        )
        self.buffer.write(self.__class__.report_messages["WHITESPACE"])
        self.buffer.write(
            self.__class__.report_messages[
                "NUM_CHECKS_EVALUATED_BOLD"].format(df.shape[0])
        )
        self.buffer.write(self.__class__.report_messages["WHITESPACE"])

        if not failing.empty:
            self.buffer.write(
                self.__class__.report_messages[
                    "NUM_FAILING_CHECKS_BOLD"].format(
                    failing.shape[0], 100 * failing.shape[0] / df.shape[0]
                )
            )
        else:
            self.buffer.write(
                self.__class__.report_messages[
                    "NUM_FAILING_CHECKS_BOLD"].format(0, 0)
            )

        if not failing.empty:
            # Number of issues of each severity
            self.buffer.write(self.__class__.report_messages["WHITESPACE"])
            self.buffer.write(self.__class__.report_messages[
                "NUMBER_SEVERITY_ISSUES"])
            self.__df_to_buffer(grouped_by_severity.reset_index(), False)

            # Number of issues of each type
            self.buffer.write(self.__class__.report_messages["WHITESPACE"])
            self.buffer.write(self.__class__.report_messages[
                "NUMBER_TYPE_ISSUES"])
            self.__df_to_buffer(grouped_by_check_type.reset_index(), False)

            # Number of issues in each data file
            self.buffer.write(self.__class__.report_messages["WHITESPACE"])
            self.buffer.write(self.__class__.report_messages[
                "NUMBER_DATA_ORIGIN"])
            self.__df_to_buffer(group_by_data_origin.reset_index(), False)

            # Number of issues of each severity for each parameter
            self.buffer.write(
                self.__class__.report_messages[
                    "NUMBER_ISSUES_PARAMETER_BY_TYPE"]
            )
            self.__df_to_buffer(grouped_by_var_name_and_severity)

            # Number of issues per parameter
            self.buffer.write(self.__class__.report_messages["WHITESPACE"])
            self.buffer.write(
                self.__class__.report_messages["NUMBER_ISSUES_PARAMETER"]
            )
            self.__df_to_buffer(grouped_by_param_name)
            self.buffer.write(self.__class__.report_messages["WHITESPACE"])

        # Passing tests
        self.buffer.write(self.__class__.report_messages["WHITESPACE"])
        self.buffer.write(self.__class__.report_messages["PASSING_HEADER"])

        if not passing.empty:
            passing_simplified = passing[
                self.__class__.default_report_columns_passing].replace(
                    {"check_name": self.__class__.check_names_map}
                    ).rename(
                        columns=self.__class__.column_names_map
                        ).set_index(group_by)

            self.__df_to_buffer(passing_simplified.drop(
                labels="check_pass",
                axis=1))

        else:
            self.buffer.write(self.__class__.report_messages[
                "NO_PASSING_CHECKS"])

    def __df_to_buffer(self, df, index=True):
        """
        Writes DataFrame to buffer.

        Parameters
        -----------
        df: pd.DataFrame
            DataFrame to write to buffer.

        index: bool
            If True, index will be printed.

        Returns
        --------
        None

        """
        self.buffer.write("\n")

        if index:
            df.index.names = [None]*len(df.index.names)
            self.buffer.write(df.to_html(justify="left"))

        else:
            self.buffer.write(df.to_html(justify="left", index=index))

        self.buffer.write("\n\n")



"""
SDQC loading file
"""
import warnings
import logging
import configparser
from ast import literal_eval
from pathlib import Path
import pysd
from pysd.builders.python.namespace import NamespaceManager
from pysd.builders.python.subscripts import SubscriptManager
from pysd.translators.structures.abstract_model import AbstractSubscriptRange

from .checks_dict import checks_dict as checks

# keep missing values of external objects
pysd.external.External.missing = "keep"


def load_data(file_name, elements_subset=None, files_subset=None):
    """
    Loads the data.

    Parameters
    ----------
    file_name: str or pathlib.Path
        File name to load external data from. It must be a Vensim model file
        (.mdl) or python PySD compatible file (.py).

    elements_subset: list
        List of elements to load. If None, all elements are loaded. It can be
        a list of original_names, py_name or py_short_name.

    files_subset: list
        List of files to load. If None, all files are loaded. File paths must
        be relative to the model file.

    Returns
    -------
    initialized: list
        The list of dictionaries containing the PySD External objects used in
        the model and additional information.

    """
    if isinstance(file_name, str):
        file_name = Path(file_name)

    file_extension = file_name.suffix

    if file_extension == ".mdl":
        model = pysd.read_vensim(file_name, initialize=False)
    elif file_extension == ".py":
        model = pysd.load(file_name, initialize=False)
    else:
        raise ValueError(
            "Invalid file extension.\n"
            + "The file must be Python file or Vensim's mdl file"
            )

    # Add original Vensim name information to the objects
    exts = model._external_elements
    namespace = model.namespace
    namespace.pop("Time")

    initialized = []
    for ext in exts:
        ext.original_name = None
        for element in namespace:
            if ext.py_name in ['_ext_data_' + namespace[element],
                               '_ext_constant_' + namespace[element],
                               '_ext_lookup_' + namespace[element],
                               namespace[element]]:
                # original name
                ext.original_name = element
                # short name
                ext.py_short_name = namespace[element]

        if elements_subset:  # initialize only the subset passed by the user
            if (ext.original_name.lower() in elements_subset) or \
                (ext.py_name.lower() in elements_subset) or \
                    (ext.py_short_name.lower() in elements_subset):

                __initialize_external_object(initialized, ext)

        elif files_subset:  # initialize only the variables in the files passed
            if all(map(lambda x: x in files_subset, ext.files)):
                __initialize_external_object(initialized, ext)

        else:  # initialize all model externals
            __initialize_external_object(initialized, ext)

    logging.info(f"\n{len(initialized)} external objects could be initialized "
                 f"out of {len(exts)}.\n")

    return initialized


def load_from_dict(elements, elements_subset=None, files_subset=None):
    """
    Loads external outpust from a elements dictionary.

    Parameters
    ----------
    elements: dict
        Dictionary of the elements. Check documentation #TODO add link

    elements_subset: list
        List of elements to load. If None, all elements are loaded. It can be
        a list of original_names, py_name or py_short_name.

    files_subset: list
        List of files to load. If None, all files are loaded. File paths must
        be relative to the model file.

    Returns
    -------
    initialized: list
        The list of dictionaries containing the PySD External objects used in
        the model and additional information.

    """
    namespace_obj = NamespaceManager()
    namespace = namespace_obj.namespace  # namespace of external objects
    namespace.pop("Time")
    abstract_subscripts = []  # subscript dictionary

    for element, data in elements.items():
        if data['type'] == 'SUBSCRIPT':
            # add subscripts to the dict
            abstract_subscripts.append(AbstractSubscriptRange(
                element, data['values'], []))
        elif data['type'] == 'EXTERNAL':
            # add external variables to the namespace
            namespace_obj.add_to_namespace(element)

    added = {'dataseries': {}, 'constant': {}}

    # instantiating a SubscriptManager object
    subscript_obj = SubscriptManager(abstract_subscripts, None)

    for element in namespace:

        if len(elements[element]['excel']) == 1:
            final_coords = subscript_obj.make_coord_dict(
                    elements[element]['excel'][0]["subs"] or [])
        else:
            final_coords = subscript_obj.make_coord_dict(
                subscript_obj.make_merge_list(
                 [ex["subs"] for ex in elements[element]['excel']]))

        for excel in elements[element]['excel']:
            # retrieve excel information
            filename = excel.get('filename')
            sheet = excel.get('sheet')
            cell = excel.get('cell')
            # TODO errors for missing filename/sheet/cell
            root = Path(excel.get('root'))
            x_row_or_col = excel.get('x_row_or_cols')
            coords = subscript_obj.make_coord_dict(excel["subs"] or [])

            if x_row_or_col:
                # convert series data to pysdexternal Lookup object
                if element in added['dataseries']:
                    # add table to existing object
                    added['dataseries'][element].add(
                        filename, sheet, x_row_or_col, cell, coords)
                else:
                    # create new object
                    obj = pysd.external.ExtLookup(
                        filename, sheet, x_row_or_col, cell, coords,
                        root, final_coords,
                        '_ext_lookup_'+namespace[element])
                    obj.original_name = element
                    obj.py_short_name = namespace[element]
                    added['dataseries'][element] = obj
            else:
                # convert constant data to pysdexternal Constant object
                if element in added['constant']:
                    # add table to existing object
                    added['constant'][element].add(
                        filename, sheet, cell, coords)
                else:
                    # TODO check the final_coords value
                    # create new object
                    obj = pysd.external.ExtConstant(
                        filename, sheet, cell, coords,
                        root, final_coords,
                        '_ext_constant_'+namespace[element])
                    obj.original_name = element
                    obj.py_short_name = namespace[element]
                    added['constant'][element] = obj

    exts_ = list(added['dataseries'].values())\
        + list(added['constant'].values())

    initialized = []
    for ext in exts_:
        if elements_subset:  # initialize only the subset passed by the user
            if (ext.original_name.lower() in elements_subset) or \
                (ext.py_name.lower() in elements_subset) or \
                    (ext.py_short_name.lower() in elements_subset):

                __initialize_external_object(initialized, ext)

        elif files_subset:  # initialize only the variables in the files passed
            if all(map(lambda x: x in files_subset, ext.files)):
                __initialize_external_object(initialized, ext)

        else:  # initialize all model externals
            __initialize_external_object(initialized, ext)

    return initialized


def load_config(config_file):
    """
    Load default configuration and user configuration if any.

    Parameters
    ----------
    config_file: str or pathlib.Path
      Path to the user configuration file. If None, no file is loaded.

    Returns
    -------
    config_dict: dict
      Dictionary with the configuration retrieved from the config file

    """
    # dictionary of arguments to load from config file and their type
    config_args = {}
    for check in checks.values():
        config_args[check['flag']] = 'bool'
        if check.get('args', None):
            for arg, (type, arg_in_func) in check['args'].items():
                config_args[arg] = type

    # TODO let define groups
    # TODO enable checks by dimension/subscript range

    # load default configuration
    config = configparser.ConfigParser()
    config.read(Path(__file__).parent / 'default-conf.ini')
    config_dict = {'all': {}, 'dataseries': {}, 'constant': {}}

    # update default configuration with user configuration
    if config_file:
        if isinstance(config_file, str):
            config_file = Path(config_file)
        if not config_file.is_file():
            raise FileNotFoundError(
                f"No such file or directory: '{config_file}'")
        # read user's config file
        config.read(config_file)

    for section in config.keys():
        sec = section.lower()
        if sec not in config_dict:
            config_dict[sec] = {}
        for key in config[section].keys():
            # convert default values to a dictionary
            if key not in config_args:
                warnings.warn(f"{key} defined in the config file is not"
                              " used in any check")
            elif config_args[key] == 'bool':
                # Boolean value (bool)
                config_dict[sec][key] = config.getboolean(section, key)
            elif config_args[key] == 'float':
                # Numeric value (float)
                config_dict[sec][key] = config.getfloat(section, key)
            elif config_args[key] == 'int':
                # Integer value (int)
                config_dict[sec][key] = config.getint(section, key)
            elif config_args[key] == 'str':
                # Type value (str)
                config_dict[sec][key] = config.get(section, key)
            elif config_args[key] == 'list (float)':
                # Range value (list)
                value = config.get(section, key)
                if value[0] == '[':
                    # literla evaluation of a list
                    config_dict[sec][key] = literal_eval(value)
                else:
                    value = [float(val) for val in value.split()]
                    if all([val.is_integer() for val in value]):
                        # if all are integers convert them to int class
                        value = [int(val) for val in value]

                    config_dict[sec][key] = value
            else:
                raise ValueError(
                    f"Invalid value type definde in checks_dict for {key}")

    return config_dict


def __initialize_external_object(initialized, ext):
    """
    Initialize external objects and add them to the initialized list.

    Parameters
    ----------
    initialized: list
        List of initialized objects.
    ext: pysd.py_backend.external.External

    """
    try:
        ext.initialize()
        add_info(ext)
        initialized.append(ext)
    except Exception as err:
        warnings.warn(
            err.args[0]
            + "\n\nNot able to initialize the following object:\n"
            f"{ext.py_name}\nFind error details above.\n")


def add_info(ext):
    """
    Add info dict to external to accelerate the DataFrame generation.

    Parameters
    ----------
    ext: pysd.py_backend.external.External

    """
    if not hasattr(ext, "py_short_name"):
        ext.py_short_name = ext.py_name
    if not hasattr(ext, "original_name"):
        ext.original_name = ext.py_name

    ext.info = {
            'py_name': ext.py_name,
            'py_short_name': ext.py_short_name,
            'original_name': ext.original_name,
            'file': ext.files,
            'sheet': ext.sheets,
            'transposed': getattr(ext, "transposed", None),
            'cell': ext.cells,
            'coords': ext.coordss,
            'final_coords': ext.final_coords
    }

    if hasattr(ext, 'time_row_or_cols'):
        # ExtLookup
        ext.info['x_row_or_col'] = ext.time_row_or_cols
    elif hasattr(ext, 'x_row_or_cols'):
        # ExtData
        ext.info['x_row_or_col'] = ext.x_row_or_cols



"""
SDQC main file
"""
import warnings
from pathlib import Path

import numpy as np
import pandas as pd
from pysd.py_backend.external import External

from .loading import load_data, load_config, load_from_dict, add_info
from .tools import _add_config_to_ext
from .reports import Report
from .checks_dict import checks_dict as checks


def check(source, config_file=None, output='dataframe',
          report_config=None, report_filename=None, verbose=False,
          elements_subset=None, files_subset=None):
    """
    Main function to run the check.

    Parameters
    ----------
    source: str, pathlib.Path, dict or list
        File name or dictionary to load external data from or list of
        dictionaries containing initialized External objects in the obj key.
        If a file name is given, it must be a Vensim model file
        (.mdl) or python PySD compatible file (.py). If a dict is
        given it should be compatible with loading.load_from_dict
        function (see #TODO link to documentation).

    config_file: str (optional)
        Path to the config file.

    output: str (optional)
        Options: "dataframe" or "report". When "report" is chosen, the user can
        define the type of report using the report_config file. Supported
        report formats are html, markdown and pdf.

    report_config: str or pathlib.Path (optional)
        Path to the config file of the report.

    report_filename: str or pathlib.Path (optional)
        Name of the report file, including extension.

    verbose: bool (optional)
        If True the output from all checks is saved. If False (default)
        only the output from not passed checks will be saved.

    elements_subset: list (optional)
        List of model variables that load external data that the user wants to
        check. This prevents the loading of all external data of the model.

    files_subset: list (optional)
        List of files that the user wants to check. This prevents the loading
        of external data from the rest of files. File paths must be relative to
        the model file. The elements_subset argument prevails if both are
        passed.

    Returns
    -------
    out: pd.DataFrame or Report
        Either a dataframe with the results of the checks or a Report object.

    """

    if isinstance(source, (str, Path)):
        # load external objects from .mdl/PySD model in Python
        exts = load_data(source, elements_subset, files_subset)
    elif isinstance(source, dict):
        # load data from a dictionary
        exts = load_from_dict(source, elements_subset, files_subset)
    elif isinstance(source, list) and all(
         map(lambda y: isinstance(y, External), source)):
        # use the list of initialized External objects
        exts = source
        [add_info(ext) for ext in exts]
    else:
        raise ValueError("Source must be a file_name, a Path, a dictionary "
                         "or a list of initialized pysd External objects.")

    # load configuration
    config = load_config(config_file)

    # add config information to each external object

    _add_config_to_ext(exts, config)

    # iterate over External objects to run the checks
    out = []
    for ext in exts:
        if ext.check_config['_type'] == 'constant':
            check_constant(ext, out, verbose)
        elif ext.check_config['_type'] == 'dataseries':
            check_dataseries(ext, out, verbose)

    if output == "dataframe":
        return pd.DataFrame(out)
    elif output == "report":
        report_obj = Report(out)
        if not report_obj.data.empty:  # if there is data to report
            report = report_obj.build_report(report_config, verbose)
            report_obj.report_to_file(report, report_filename)

        return report_obj
    else:
        raise ValueError("Output must be 'dataframe' or 'html'.")


def check_constant(ext, out, verbose):
    """
    Run the checks in a ExtConstant type object.

    Parameters
    ----------
    ext: dict
        The dictionary contains the pysd.external.ExtConstant object to pass
        the check in the obj key. It must have the py_short_name, original_name
        and config attributes, set previously.
    out: list
        The list to append the results of the checks.
    verbose: bool
        If True the output from all checks is saved. If False only the
        output from not passed checks will be saved.

    Returns
    -------
    None

    """
    # TODO enable checks by dimension/subscript range

    if not hasattr(ext, 'data'):
        return

    if ext.coordss[0]:
        constant_values = ext().values
    else:
        # if constant value is 0-D (float) no test can be passed
        return

    # Start checks
    for chk in checks:
        # Is this check type selected and can it be run on constants?
        if 'constant' in checks[chk]['targets']\
          and ext.check_config[checks[chk]['flag']]:
            new_info = _run_check(
                [constant_values], chk, ext.check_config,
                'constant', ext.info, verbose)
            if new_info:
                out.append(new_info)


def check_dataseries(ext, out, verbose):
    """
    Run the checks in a ExtData or ExtLookup type object.

    Parameters
    ----------
    ext: dict
        The dictionary contains the pysd.external.ExtData or
        pysd.external.ExtLookup object to pass the check in the obj key. It
        must have the py_short_name, original_name and config attributes,
        set previously.
    out: list
        The list to append the results of the checks.
    verbose: bool
        If True the output from all checks is saved. If False only the
        output from not passed checks will be saved.

    Returns
    -------
    None

    """
    # TODO enable checks by dimension/subscript range

    if not hasattr(ext, "data"):
        return

    data_values = ext.data.values
    series_values = ext.data.coords[ext.data.dims[0]].values

    if ext.check_config["missing"]:
        # check for missing values in the series first, and add them to the
        # result of the check, as we need to remove them before starting the
        # other checks
        new_info = ext.info.copy()
        new_info['check_name'] = 'missing_values_series'
        new_info['check_description'] = checks['missing_values']['descr']
        new_info['check_target'] = 'series'
        new_info['check_arg'] = {}

        # run the check for missing values in the series
        new_info['check_pass'], new_info['check_out'] =\
            checks['missing_values']['func'](series_values)

        if verbose or not new_info['check_pass']:
            out.append(new_info)

    # here the user did not select that they want to check for missing values,
    # but we are removing them anyway
    elif np.any(np.isnan(series_values)):
        warnings.warn(
            "Removing the columns corresponding to missing values in "
            "the series. Activate the missing_values check to see "
            "the results of the check.")

    # Remove missing values from series, if any
    data_values = data_values[~np.isnan(series_values)]
    series_values = series_values[~np.isnan(series_values)]

    # Start checks
    for chk in checks:
        # Is the check activated?
        new_info = None
        if ext.check_config[checks[chk]['flag']]:
            if 'series' in checks[chk]['targets']:
                # Check for series
                new_info = _run_check([series_values], chk,
                                      ext.check_config, 'series', ext.info,
                                      verbose)
            elif 'data' in checks[chk]['targets']:
                # Check for data
                new_info = _run_check([data_values], chk,
                                      ext.check_config,
                                      'data', ext.info, verbose)
            elif 'dataseries' in checks[chk]['targets']:
                # Check for data + series
                new_info = _run_check([data_values, series_values], chk,
                                      ext.check_config, 'dataseries',
                                      ext.info, verbose)
            if new_info:
                out.append(new_info)


def _run_check(values, check_name, econfig, target, info, verbose):
    """
    Run a check with the given arguments.

    Parameters
    ----------
    values: list
        List of arguments to run a check, could be data, series,
        constant or data and series.
    check_name: str
        Name of the check to be passed.
    econfig: dict
        Configuration dictionary of the object. Needed to find extra
        arguments for the function.
    target: str
        The target of the check, must match the arguments passed in args.
    info: dict
        General info for the current reading block.
    verbose: bool
        If True the output from all checks is saved. If False only the
        output from not passed checks will be saved.

    Returns
    -------
    None

    """
    # create a new dictionary to sabe
    new_info = info.copy()

    # save name, description and target
    new_info['check_name'] = check_name
    new_info['check_description'] = checks[check_name]['descr']
    new_info['check_target'] = target

    if 'args' in checks[check_name] and checks[check_name]['args']:
        # read the arguments and save the info
        new_info['check_arg'] = {
            checks[check_name]['args'][arg][1]: econfig[arg]
            for arg in checks[check_name]['args']}
    else:
        # no arguments
        new_info['check_arg'] = {}

    # run the check
    new_info['check_pass'], new_info['check_out'] =\
        checks[check_name]['func'](*values,
                                   **new_info['check_arg'],
                                   coords=new_info['coords'],
                                   final_coords=new_info['final_coords'])

    # save the result if the check has not passed or verbose is True
    if not new_info['check_pass'] or verbose:
        return new_info



"""
SDQC tools file
"""
from pysd import external as pysdexternal


def _add_config_to_ext(exts, config):
    """
    Add checks configuration information to each External object.

    Parameters
    ----------
    exts: list
        The list of dictionaries that contain the pysd.external.External
        objects used in the model in the obj key.

    Returns
    -------
    None

    """
    for ext in exts:
        # TODO enable checks by dimension/subscript range
        # Include parameters for All
        extconfig = config['all'].copy()

        # Check if the object is Constant or Data and include specific params
        if isinstance(ext, pysdexternal.ExtConstant):
            extconfig.update(config['constant'])
            extconfig['_type'] = 'constant'
        elif isinstance(ext, (pysdexternal.ExtData, pysdexternal.ExtLookup)):
            extconfig.update(config['dataseries'])
            extconfig['_type'] = 'dataseries'

        # TODO add group option with level between object name and type
        # Find if the object is in the dictionary
        if ext.original_name.lower() in config:
            extconfig.update(config[ext.original_name.lower()])
        elif ext.py_name.lower() in config:
            extconfig.update(config[ext.py_name.lower()])
        elif ext.py_short_name.lower() in config:
            extconfig.update(config[ext.py_short_name.lower()])

        # set the attribute
        ext.check_config = extconfig



from . import checks as c

checks_dict = {
    # name_of_the_check: {
    #    'func': check_function,
    #    'descr': description,
    #    'flag': name_of_the_flag_in config_file_to_activate_check,
    #    'targets': [targets ('constant', 'data', 'series', 'dataseries')]
    #    'args': {
    #       name_of_the_arg_in config_file: (type, name_of_the_arg_in_function)
    #       }
    # }
    'missing_values': {
        'func': c.missing_values,
        'descr': 'Checks for missing values on the data.',
        'flag': 'missing',
        'targets': ['constant']
    },
    'missing_values_data': {
        'func': c.missing_values_data,
        'descr': 'Checks for missing values on the data giving the series.',
        'flag': 'missing',
        'targets': ['dataseries'],
        'args': {'completeness': ('str', 'completeness')}
    },
    'series_monotony': {
        'func': c.series_monotony,
        'descr': 'Checks if series is monotonically increasing.',
        'flag': 'series_monotony',
        'targets': ['series']
    },
    'series_range': {
        'func': c.series_range,
        'descr': 'Checks if series isinside a range.',
        'flag': 'series_range',
        'targets': ['series'],
        'args': {'series_range_values': ('list (float)', 'srange')}
    },
    'series_increment_type': {
        'func': c.series_increment_type,
        'descr': 'Checks if series series increment type.',
        'flag': 'series_increment',
        'targets': ['series'],
        'args': {'series_increment_type': ('str', 'type')}
    },
    'outlier_values': {
        'func': c.outlier_values,
        'descr': 'Checks for outlier values on the data.',
        'flag': 'outliers',
        'targets': ['constant', 'data'],
        'args': {'outliers_method': ('str', 'method'),
                 'outliers_nstd': ('float', 'nstd'),
                 'outliers_niqr': ('float', 'niqr')}
    }
}



"""
SDQC checks file
"""
import numpy as np


def missing_values(data, **kwargs):
    """
    Checks for missing values on the data.

    Parameters
    ----------
    data: ndarray
        The data to check.

    Returns
    -------
    pass, outs: bool, list
        pass is True if no missing values have been found and False if
        missing values have been found. outs is a list with the number
        of missing values and the total number of values.

    """
    n_nan = int(np.sum(np.isnan(data)))
    n_values = int(np.prod(data.shape))

    if n_nan == 0:
        return True, [0, n_values]

    return False, [n_nan, n_values]


def missing_values_data(data, series, completeness, **kwargs):
    """
    Checks for missing values on the data giving the series.

    Parameters
    ----------
    data: ndarray
        The data to check.
    series: ndarray (1D)
        The series to check.
    completeness: str (optional)
        If completeness is 'any' (default) the check will not pass
        if there is any missing value for a given series value (column).
        If completeness is 'all' the check will not pass if all the data
        values are missing for a given series value .
        It only has effects when data is a matrix (2 or more dimensions).

    Returns
    -------
    pass, outs: bool, list
        pass is True if no missing values have been found and False if
        missing values have been found. outs is a list with a first list
        of the number of missing values and the total number of values
        checked, and a second list with the series values for which the
        missing values where found. Note that the number of missing
        values and the total values depens on the completeness.

    """
    data_nan = np.isnan(data)
    n_nan = int(np.sum(data_nan))
    n_values = int(np.prod(data_nan.shape))

    if len(data.shape) >= 2:
        if completeness == 'all':
            data_nan = np.all(data_nan, axis=tuple(range(1, data_nan.ndim)))
            n_nan = int(np.sum(data_nan))
            n_values = int(np.prod(data_nan.shape))
        elif completeness == 'any':
            data_nan = np.any(data_nan, axis=tuple(range(1, data_nan.ndim)))
        else:
            raise ValueError("Invalid value for completeness argument.")

    if not any(data_nan):
        return True, [[0, n_values], []]

    missing = list(series[data_nan])
    return False, [[n_nan, n_values], missing]


def series_monotony(series, **kwargs):
    """
    Checks if series is monotonous.

    Parameters
    ----------
    series: ndarray (1D)
        The series to check.

    Returns
    -------
    pass, outs: bool, list
        pass is True if series is monotonous and False if not.
        outs is a list with the series range (minimum and maximum).

    """
    if all(np.diff(series) < 0) or all(np.diff(series) > 0):
        return True, [series[0], series[-1]]

    return False, [np.min(series), np.max(series)]


def series_range(series, srange, **kwargs):
    """
    Checks if series range is inside a given range.

    Parameters
    ----------
    series: ndarray (1D)
        The series to check.

    srange: list (len 2)
        The minimum and maximum value of the series.

    Returns
    -------
    pass, outs: bool, list
        pass is True if series is monotonous and False if not.
        outs is a list with the series range (minimum and maximum)
        and a list with the give srange.

    """
    smin, smax = np.min(series), np.max(series)
    if smin < srange[0] or smax > srange[1]:
        return False, [[smin, smax], srange]

    return True, [[smin, smax], srange]


def series_increment_type(series, type, **kwargs):
    """
    Checks if series increments following the trend specified by the type
    argument.

    Parameters
    ----------
    series: ndarray (1D)
        The series to check.

    type: str
        The series distribution ('linear')

    Returns
    -------
    pass, outs: bool, list
        pass is True if series is monotonous and False if not.
        outs is a list with the series step or the full series.

    """
    # TODO include more distributions ('exponential', 'logarithmic'...)
    if type == 'linear':
        ds = np.diff(series)
        if np.all(ds == ds[0]):
            # return True with the step
            return True, [ds[0]]
    else:
        raise ValueError(
            f"The argument {type} is not supported by series_increment_type.")

    return False, list(series)


def outlier_values(data, method, **kwargs):
    """
    Checks if there are possible outliers in the data using a
    multiple of the standard deviation.

    Parameters
    ----------
    data: ndarray
        The data to check.

    method: str
         The method to be used. Can be 'std' for standard deviation method
         or 'iqr' for interquartile range method.

    **kwargs: see below

    Keyword Args
    ------------
    nstd: float
        For 'std' method, the number of standard deviations
        to define outliers.

    niqr: float
        For 'iqr' method, the number of interquartile ranges
        to define outliers.

    Returns
    -------
    pass, outs: bool, list
        pass is True if the data does not have outliers and False if it does.
        outs is a list with the number of possible outliers and
        the total number of values.
    """
    # number of values in the ndarray
    n_values = data.size

    if method == 'std':
        # standard deviation method
        nstd = kwargs.get('nstd', 2)
        mean = np.nanmean(data)
        std = np.nanstd(data)
        # mark outlier values as True
        mask = np.logical_or(data < mean - nstd*std,
                             data > mean + nstd*std)

    elif method == 'iqr':
        # interquartile range method
        niqr = kwargs.get('niqr', 1.5)
        q_low, q_up = np.percentile(data, 25), np.percentile(data, 75)
        iqr = q_up - q_low
        # mark outlier values as True
        mask = np.logical_or(data < q_low - niqr*iqr,
                             data > q_up + niqr*iqr)

    else:
        raise ValueError(
            f"The method {method} is not supported by outlier_values."
        )

    # no outliers found
    if not np.any(mask):
        return True, [[], n_values]

    coords = kwargs.get('coords', [])
    final_coords = kwargs.get('final_coords', {})

    # get indexes of outlier values. This returns as many arrays as there are
    # dimensions in the ndarray. Each array contains the indexes of the outlier
    # in that dimension.
    outlier_indexes = np.where(mask)

    # If the data has dimensions the outlier_indexes is a list of as many
    # arrays as there are dimensions in the ndarray. Here we make lists with
    # the indexes of the outlier values in each dimension.
    if len(data.shape) > 1:
        outlier_indexes = [list(idxs) for idxs in zip(*outlier_indexes)]

    outliers = []
    if len(coords) > 1:  # data loaded from multiple files/tabs/cellrange names

        # here we classify outlier values in separate lists, corresponding to
        # the file/tab/cellrange name from which the values were loaded

        # join all unique coordinates of all dimensions from the different
        # files/tabs/cellrange names into a single list
        grouped_coords = [set().union(*element.values()) for element in coords]

        # put the coordinates of each dimension of the final_coords in a
        # separate list
        grouped_final_coords = list(final_coords.values())

        # number of values in the ndarray obtained from dimensions.
        # when we compare this with the number of values in the data, we will
        # and they don't match, we will know the data has a time dimenision.
        n_values_from_coords = 1
        for x in grouped_final_coords:
            n_values_from_coords *= len(x)

        for el in grouped_coords:
            sublist = []
            for idx in outlier_indexes:  # idx should be smthng like [4, 2, 1]

                # removing the time dimension, if it exists
                idx_ = idx[1:] if n_values != n_values_from_coords else idx

                # get the names of the coordinates corresponding to the
                # indexes of the identified outliers
                coord_names = []
                for i in range(len(idx_)):
                    coord_names.append(grouped_final_coords[i][idx_[i]])

                # if those coordinate names are all in the coords of the
                # specific file/tab/cellrange name we are iterating over (el),
                # it means that the outlier is in that file/tab/cellrange,
                # so we add it to a separate list for that particular
                # file/tab/cellrange name.
                if all(map(lambda x: x in el, coord_names)):
                    sublist.append(data[tuple(idx)].item())

            outliers.append(sublist)

    else:
        # data loaded from a single file/tab/cellrange name,
        # hence we put all outliers in a single list
        if len(data.shape) > 1:  # has dimensions
            outliers = [data[tuple(idx)].item() for idx in outlier_indexes]
        else:  # without dimensions
            outliers = [data[idx].item() for idx in outlier_indexes[0]]

    return False, [outliers, n_values]
